language = 'spanish'
iso639 = 'es'

spacy_model_sm = iso639 + '_core_news_sm'
spacy_model_md = iso639 + '_core_news_md'
spacy_model_lg = iso639 + '_core_news_lg'

sentiment_lexicon_class = None
pos_token_sentiment = None

negations = ['nadie', 'nada', 'ni', 'jamás', 'jamas', 'nunca', 'tampoco', 'todavía no', 'todavia no',
                    'ya no', 'no', 'ninguno', 'ningún', 'ningun', 'ni siquiera', 'más']