language = 'chinese'
iso639 = 'zh'

spacy_model_sm = iso639 + '_core_news_sm'
spacy_model_md = iso639 + '_core_news_md'
spacy_model_lg = iso639 + '_core_news_lg'

sentiment_lexicon_class = 'Fannix'
pos_token_sentiment = ['AUX', 'VERB', 'ADJ', 'ADV', 'NOUN', 'PROPN']

negations = ['bù', '不', 'méi', '没', 'méiyǒu', '没有', "甚至", "完全不", "少", "決不", "不再"]
