language = 'greek'
iso639 = 'el'

spacy_model_sm = iso639 + '_core_news_sm'
spacy_model_md = iso639 + '_core_news_md'
spacy_model_lg = iso639 + '_core_news_lg'

sentiment_lexicon_class = None
pos_token_sentiment = None

negations = ['όχι', 'κανένας', 'ούτε', 'ποτέ', 'δεν', 'κανένας', 'νιξ', 'μη ουσία', 'ακυρότητα', 'τίποτα']