language = 'dutch'
iso639 = 'nl'

spacy_model_sm = iso639 + '_core_news_sm'
spacy_model_md = iso639 + '_core_news_md'
spacy_model_lg = iso639 + '_core_news_lg'

sentiment_lexicon_class = None
pos_token_sentiment = None

negations = ['zelfs niet', 'zelfs niet', 'zelfs niet', 'niemand', 'niemand', 'niets', 'niets', 'helemaal niet',
             'niet', 'minder', 'nooit meer', 'nooit', 'nee']