language = 'danish'
iso639 = 'da'

spacy_model_sm = iso639 + '_core_news_sm'
spacy_model_md = iso639 + '_core_news_md'
spacy_model_lg = iso639 + '_core_news_lg'

sentiment_lexicon_class = None
pos_token_sentiment = None

negations = ['ikke engang', 'ikke en gang', 'ingen', 'nogen', 'ingen', 'intet', 'slet ikke', 'ikke', 'mindre',
             'aldrig igen', 'aldrig', 'nej']