__all__ = ['fannix', 'sentiwordnet', 'vader', 'sentix']

from compling.lexicons.vader import Vader
from compling.lexicons.fannix import Fannix
from compling.lexicons.sentix import Sentix
from compling.lexicons.sentiwordnet import Sentiwordnet