language = 'german'
iso639 = 'de'

spacy_model_sm = iso639 + '_core_news_sm'
spacy_model_md = iso639 + '_core_news_md'
spacy_model_lg = iso639 + '_core_news_lg'

sentiment_lexicon_class = None
pos_token_sentiment = None

negations = ['nicht', 'nein', 'es ist nicht', 'weder', 'noch', 'noch nie', 'kaum', 'selten', 'ist nicht',
       'sollte nicht', 'konnte nicht', 'wird nicht', 'kann nicht ', 'niemand', 'nichts', 'nichtigkeit']