language = 'norwegian-bokmål'
iso639 = 'nb'

spacy_model_sm = iso639 + '_core_news_sm'
spacy_model_md = iso639 + '_core_news_md'
spacy_model_lg = iso639 + '_core_news_lg'

sentiment_lexicon_class = None
pos_token_sentiment = None

negations = ['ikke', 'ikke', 'nei', 'ingen', 'verken', 'eller', 'aldri', 'ingen', 'intet', 'null', 'nullitet',
       'ingenting']